import time
import sys
from peucr_core.loaders import ConfigLoader, SpecLoader, PluginLoader, ValidatorLoader
from peucr_core.exceptions import InvalidDefinitionException


class TestFramework:
    def __init__(self, args):
        self.separator = "**************************************************"
        self.retryInterval = 0.2
        self.config = ConfigLoader(args).apply()
        self.specs = SpecLoader(self.config).apply()
        self.plugins = PluginLoader(self.config).apply()
        self.validators = ValidatorLoader(self.config).apply()



    def exec_actions(self, actions):
        if actions is None or len(actions) == 0:
            return True

        for action in actions:
            executed = False
            startTime = time.time()

            while not executed and time.time() - startTime < 2:
                try:
                    r = self.plugins.apply(action)
                    executed = r["success"]
                except Exception as e:
                    pass

                if executed:
                    break

                time.sleep(self.retryInterval)

        if not executed:
            print("Action failed. Validation will not be executed")
            return False

        return True



    def exec_validation(self, validation):
        time.sleep(validation.get("wait", 0))

        attempts = min(5, validation.get("duration", self.retryInterval)) / self.retryInterval
        counter = 0
        result = {"success": False}

        while not result["success"] and counter < attempts:
            try:
                response = self.plugins.apply(validation)
                result = self.validators.apply(validation.get("expectation"), response)

            except InvalidDefinitionException as e:
                result["msg"] = e
                break

            except Exception as e:
                result["msg"] = "Error:", e

            if result["success"]:
                break

            time.sleep(self.retryInterval)
            counter += 1

        if not result["success"]:
            print("Failure.", result.get("msg", ""))

        return result["success"]



    def exec_validations(self, validations):
        if not validations or not isinstance(validations, list) or len(validations) == 0:
            print("No validation specified in test. Aborting.")
            return False

        for validation in validations:
            if not self.exec_validation(validation):
                return False

        return True



    def exec_test(self, spec):
        print("Executing test \"{}\"".format(spec.get("name", "UNNAMED")))

        if not self.exec_actions(spec.get("context")):
            return False

        if not self.exec_actions(spec.get("actions")):
            return False

        return self.exec_validations(spec.get("validation"))



    def exec_test_suite(self, specs):
        successes = 0

        for spec in specs:
            if self.exec_test(spec):
                successes += 1

        print(self.separator)

        if successes != len(specs):
            print(len(specs), "tests run", len(specs) - successes, "failures")
            sys.exit(1)        

        print(len(specs), "tests run. No failures.")



    def exec_preconditions(self, validations):
        if validations is None or validations.get("validation") is None:
            return

        for validation in validations["validation"]:
            print("Verifying", validation.get("name", "UNNAMED"))
            if not self.exec_validation(validation):
                print("Precondition validation failed. Test will be aborted")
                sys.exit(1)

        print(self.separator)



    def exec(self):
        self.exec_preconditions(self.specs.get("preconditions"))
        self.exec_test_suite(self.specs.get("execution"))



import sys
import os
import inspect
import importlib
import json
from peucr_core.validatorsuite import ValidatorSuite
from peucr_core.pluginsuite import PluginSuite


class ConfigLoader:

    def __init__(self, args):
        self.prefix = "--"
        self.args = args


    def apply(self):
        config = {}

        if not self.get(self.args, "plugins"):
            config["plugins"] = "plugins"

        if not self.get(self.args, "specs"):
            config["specs"] = "specs"

        if not self.get(self.args, "validators"):
            config["validators"] = "validators"

        for index in range(len(self.args)):
            if self.is_flag(self.args, index):
                value = self.get(self.args, self.args[index])
                if value:
                    config[self.args[index][len(self.prefix):]] = value

        return config


    def get(self, args, name):
        flag = self.prefix + name if name[:2] != self.prefix else name

        if flag not in args:
            return None

        index = args.index(flag)

        if len(args) <= index+1 or self.is_flag(args, index+1):
            return None

        return args[index+1]


    def is_flag(self, args, index):
        return self.prefix == args[index][:2]



class PluginLoader:

    def __init__(self, config):
        self.plugin = "TestPlugin"
        self.config = config


    def apply(self):
        path = self.config["plugins"]

        if path[-1] == "/":
            path = path[0:-1]
        parentModule = path.split("/")[-1]

        plugins = []

        if os.path.exists(path):
            for file in os.listdir(path):
                try:
                    lib = importlib.import_module(parentModule+"."+file.split(".")[0])
                except Exception as e:
                    print(e)
                    sys.exit(1)
                    
                members = inspect.getmembers(lib, inspect.isclass)
                
                plugins.append(self.getTestPlugin(members, parentModule, lib))
            
        return PluginSuite([plugin(self.config) for plugin in plugins if plugin is not None], self.config)


    def getTestPlugin(self, member, parent, lib):
        pluginClass = None
        testPlugin = False

        for clazz in member:
            if clazz[0] == self.plugin:
                testPlugin = True

            if parent+"." in str(clazz[1]):
                pluginClass = clazz[0]

        if testPlugin and pluginClass:
            return getattr(lib, pluginClass)



class SpecLoader:

    def __init__(self, config):
        self.spec = "_spec.json"
        self.preconditions = "preconditions.json"
        self.config = config

    def apply(self):
        path = self.config["specs"]

        specs = {}
        if self.spec in path:
            directory = "/".join(path.split("/")[:-1])
            files = os.listdir(directory)
            spec_files = [path.split("/")[-1]]
        else:
            directory = path
            files = os.listdir(directory)
            spec_files = [file for file in files if "_spec.json" in file]
            spec_files.sort()
        
        preconditions = directory+"/"+self.preconditions if self.preconditions in files else None

        if preconditions:
            with open(preconditions, "r") as p:
                specs["preconditions"] = json.loads(p.read())

        execution = []
        for file in spec_files:
            with open(directory+"/"+file, "r") as f:
                execution.append(json.loads(f.read()))

        specs["execution"] = execution

        return specs



class ValidatorLoader:

    def __init__(self, config):
        self.validator = "TestValidator"
        self.config = config


    def apply(self):
        path = self.config["validators"]

        if path[-1] == "/":
            path = path[0:-1]
        parentModule = path.split("/")[-1]

        validators = []

        if os.path.exists(path):
            for file in os.listdir(path):
                try:
                    lib = importlib.import_module(parentModule+"."+file.split(".")[0])
                except Exception as e:
                    print(e)
                    sys.exit(1)
                    
                members = inspect.getmembers(lib, inspect.isclass)
                
                validators.append(self.getValidator(members, parentModule, lib))
            
        return ValidatorSuite([validator() for validator in validators if validator is not None])

    
    def getValidator(self, member, parent, lib):
        validatorClass = None
        testValidator = False

        for clazz in member:
            if clazz[0] == self.validator:
                testValidator = True

            if parent+"." in str(clazz[1]):
                validatorClass = clazz[0]

        if testValidator and validatorClass:
            return getattr(lib, validatorClass)



import requests
from peucr_core.plugin import TestPlugin

class HttpPlugin(TestPlugin):

    def __init__(self, config):
        self.labels = ["HTTP"]
        self.config = config


    def apply(self, options = {}):
        url = self.configure(options.get("url"))

        response = requests.get(url)

        success = response.status_code >= 200 and response.status_code < 300

        error = "Received status code {}".format(response.status_code) if not success else ""

        return {"success": success, "status-code": response.status_code, msg: error}



import requests
from peucr_core.validator import TestValidator
from peucr_core.exceptions import InvalidDefinitionException

class DefaultValidator(TestValidator):

    def __init__(self):
        self.labels = ["DEFAULT"]


    def apply(self, expectation, response, suite):
        if "success" not in response:
            raise InvalidDefinitionException("No \"success\" field in response")

        return {"success": response["success"]}



import requests
from peucr_core.validator import TestValidator
from peucr_core.exceptions import InvalidDefinitionException

class JsonValidator(TestValidator):

    def __init__(self):
        self.labels = ["JSON"]


    def apply(self, expectation, response, suite):
        body = response.get("json")
        if body is None:
            raise InvalidDefinitionException("No JSON body in response")

        if expectation.get("value") is None:
            raise InvalidDefinitionException("JSON expectation requires \"value\"")
        value = expectation["value"]

        if expectation.get("field") is None:
            raise InvalidDefinitionException("JSON expectation requires \"field\"")
        field = expectation["field"]

        success = value == body.get(field)
        error = "Expected \"{}\" to be \"{}\" but got \"{}\"".format(field, value, body.get(field))

        return {"success": success, "msg": None if success else error}




import re

class TestPlugin:

    def __init__(self):
        self.labels = []
        self.config = {}


    def executes(self, name):
        return name.upper() in [l.upper() for l in self.labels]


    def configure(self, value):
        pattern = r"\$\{([^}]+)\}"

        m = re.search(pattern, value)

        if m is None:
            return value

        for key in m.groups():
            if self.config.get(key):
                value = value.replace("${" + key + "}", self.config[key])
        
        return value
